"""
Ensures there are no `simulatedMetricsContainer`s, which were removed with Xcode 9
"""


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    root = context.tree.getroot()
    for container in root.findall('.//simulatedMetricsContainer'):
        context.error(container, 'SimulatedMetricsContainers should be removed (resave with Xcode 9+)')



"""
Checks for Trait Variations being enabled.
"""


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    root = context.tree.getroot()
    if root.get('useTraitCollections') == 'YES' and root.get('targetRuntime') != 'watchKit':
        context.error(root, "Document has 'Use Trait Variations' enabled")



import inspect
from glob import glob
from os.path import basename, dirname
from importlib import import_module


def _collect_checkers():
    _rule_checkers = {}
    for file in glob(dirname(__file__) + "/*.py"):
        rule_name = basename(file)[:-3]
        module = import_module('xiblint.rules.' + rule_name)
        check = getattr(module, 'check', None)
        if inspect.isfunction(check):
            _rule_checkers[rule_name] = check
    return _rule_checkers

rule_checkers = _collect_checkers()



import weakref

# HACK to ensure cElementTree is not loaded -- we need the pure Python implementation
import sys

assert 'xml.etree.ElementTree' not in sys.modules
sys.modules['_elementtree'] = {}

from xml.etree import ElementTree  # noqa E402
from defusedxml.ElementTree import DefusedXMLParser  # noqa E402


class _TreeBuilder(ElementTree.TreeBuilder):
    """
    Adds a line and column number to every element.
    """
    def __init__(self):
        super(_TreeBuilder, self).__init__()
        self.parser = None

    def start(self, tag, attrs):
        elem = super(_TreeBuilder, self).start(tag, attrs)
        elem.line = self.parser.parser.CurrentLineNumber
        elem.column = self.parser.parser.CurrentLineNumber
        return elem


def parse_xml(file_path):
    builder = _TreeBuilder()
    parser = DefusedXMLParser(target=builder, forbid_dtd=False, forbid_entities=True, forbid_external=True)
    builder.parser = weakref.proxy(parser)
    tree = ElementTree.parse(file_path, parser=parser)

    # ElementTree does not implement parent attributes: add parent property to every element
    parent_map = {child: parent
                  for parent in tree.getiterator()
                  for child in parent}
    for p in tree.getiterator():
        p.parent = parent_map.get(p)

    return tree


def element_and_parents(p):
    """
    Helper to iterate over the XML element and its parents up to root.
    """
    while p is not None:
        yield p
        p = p.parent



#!/usr/bin/env python
from xiblint import __version__
import argparse
import os
import sys

from xiblint.config import Config
from xiblint.xibcontext import XibContext
import xiblint.rules


def make_epilog_text():
    description = "Lint rules:\n"

    for rule_name in sorted(xiblint.rules.rule_checkers.keys()):
        checker = xiblint.rules.rule_checkers[rule_name]
        description += "\t{}\n".format(rule_name)
        doc = sys.modules[checker.__module__].__doc__
        if doc is not None:
            for line in doc.strip().split("\n"):
                description += " " * 24 + "{}\n".format(line)

    return description


def main():
    parser = argparse.ArgumentParser(
        description='.xib / .storyboard linter',
        epilog=make_epilog_text(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("-v", "--version", action="version",
                        version=__version__)
    parser.parse_args()

    try:
        config = Config()
    except IOError as ex:
        print('Error: {}\n'.format(ex))
        parser.print_usage()
        sys.exit(1)
    except ValueError as ex:
        print('Error: {}: {}\n'.format(ex, Config.filename, ex))
        parser.print_usage()
        sys.exit(1)

    #
    # Process paths
    #
    success = True
    for root, _, files in os.walk(u'.'):
        for file_path in [os.path.join(root, file) for file in files]:
            checkers = config.checkers(file_path)
            success = process_file(file_path, checkers) and success

    sys.exit(0 if success else 1)


def process_file(file_path, checkers):
    _, ext = os.path.splitext(file_path)
    if ext.lower() not in [u".storyboard", u".xib"]:
        return True
    context = XibContext(file_path)
    for rule_name, checker in checkers.items():
        context.rule_name = rule_name
        checker(context)
    return context.success


if __name__ == '__main__':
    main()



import json
from os.path import abspath
from fnmatch import fnmatch

import xiblint.rules


def validate_rule_patterns(patterns):
    for pattern in (patterns or []):
        if not any(fnmatch(key, pattern) for key in xiblint.rules.rule_checkers):
            print("Warning: Rule pattern '{}' matches no rules".format(pattern))


class Config(object):
    filename = '.xiblint.json'

    def __init__(self):
        with open(self.filename, 'r') as file:
            data = json.load(file)
        self.rules = data.get('rules', [])
        validate_rule_patterns(self.rules)
        self.paths = {abspath(path): PathConfig(self, config)
                      for (path, config) in data.get('paths', {}).items()}
        self.base_config = PathConfig(self, {})

    def _config_for_file_path(self, file_path):
        file_path = abspath(file_path)
        for path, config in self.paths.items():
            if file_path == path:
                return config
            if file_path.startswith(path + '/'):
                return config
        return self.base_config

    def checkers(self, file_path):
        return self._config_for_file_path(file_path).checkers


class PathConfig(object):
    def __init__(self, config, data):
        self.config = config
        self.path_rules = data.get('rules')
        validate_rule_patterns(self.path_rules)
        self.excluded_rules = data.get('excluded_rules', [])
        validate_rule_patterns(self.excluded_rules)

        # Filter checkers
        patterns = self.path_rules if self.path_rules is not None else self.config.rules
        excluded_patterns = self.excluded_rules
        self.checkers = {
            rule: checker for (rule, checker) in xiblint.rules.rule_checkers.items()
            if any(fnmatch(rule, pattern) for pattern in patterns) and
            not any(fnmatch(rule, pattern) for pattern in excluded_patterns)
        }



"""
Checks for accessible images with no accessibility label.
In this case, VoiceOver will announce the image asset's name, which might be unwanted.
"""
from xiblint.xibutils import (
    view_is_accessibility_element,
    view_accessibility_label,
    get_view_user_defined_attr,
)


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    for image_view in context.tree.findall(".//imageView"):
        if (
            view_is_accessibility_element(image_view) is True and
            not view_accessibility_label(image_view) and
            not get_view_user_defined_attr(image_view, 'accessibilityFormat')
        ):
            context.error(image_view, "Image is accessible but has no accessibility label")



"""
Checks for labels with outlets into a view controller that have no accessibility identifiers.
Labels with outlets might get dynamic text, and therefore should be accessible to UI testing.
"""


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    for outlet in context.tree.findall(".//viewController/connections/outlet"):
        destination = outlet.get('destination')
        label = context.tree.find(".//label[@id='{}']".format(destination))
        if label is None:
            continue
        if label.find('./accessibility[@identifier]') is not None:
            continue
        context.error(label,
                      "Label with text '{}' has an outlet into the view controller "
                      "so it requires an accessibility identifier",
                      label.get('text'),
                      label.tag)



"""
Makes sure that interactive views have accessibility identifiers, to support testing through UI Automation.
"""


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    for tag in ['button', 'textField', 'textView']:
        for view in context.tree.findall(".//{}".format(tag)):
            if view.find('./accessibility[@identifier]') is None:
                context.error(view, "{} requires an accessibility identifier for UI Automation", tag)



"""
Checks for ambiguous and misplaced views.
"""


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    for element in context.tree.iter():
        if element.get('ambiguous') == 'YES':
            context.error(element, "View with ambiguous constraints")
        elif element.get('misplaced') == 'YES':
            context.error(element, "Misplaced view")



"""
Checks for text fields and text views with a placeholder and no accessibility label.
A placeholder is not a substitute for an accessibility label, since it's no longer announced
after the text is edited.
"""
from xiblint.xibutils import (
    get_view_user_defined_attr,
    view_is_accessibility_element,
)


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    for element in context.tree.findall(".//textField") + context.tree.findall(".//textView"):
        placeholder = (element.get('placeholder') if element.tag == 'textField'
                       else get_view_user_defined_attr(element, 'placeholder'))
        if (
            placeholder is not None and
            element.find('./accessibility[@label]') is None and
            view_is_accessibility_element(element) is not False
        ):
            context.error(element,
                          "{} with placeholder text '{}' but no accessibility label",
                          element.tag,
                          element.get('placeholder'))



"""
Ensures simulated metrics are for the iPhone SE, which is currently the smallest display profile.
"""


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    root = context.tree.getroot()

    # In Xcode 9 this metadata is in a new place
    device = root.find('device')
    if device is not None and device.get('id') != 'retina4_0':
        context.error(device, 'Simulated metrics ("View As:") must be set to iPhone SE.')
        return

    for simulated_screen_metrics in root.findall(".//simulatedScreenMetrics[@key='destination']"):
        if simulated_screen_metrics.get('type') != 'retina4_0.fullscreen':
            context.error(simulated_screen_metrics, 'Simulated metrics ("View As:") must be set to iPhone SE.')



"""
Checks for image buttons with no accessibility label.
In this case, VoiceOver will announce the image asset's name, which might be unwanted.
"""
from xiblint.xibutils import (
    view_is_accessibility_element,
    view_accessibility_label,
    get_view_user_defined_attr,
)


def check(context):  # type: (xiblint.xibcontext.XibContext) -> None
    for button in context.tree.findall(".//button"):
        state_normal = button.find("./state[@key='normal']")
        if (
            state_normal is None or
            'title' in state_normal.attrib or
            view_is_accessibility_element(button) is False or
            view_accessibility_label(button) or
            get_view_user_defined_attr(button, 'accessibilityFormat')
        ):
            continue

        if 'image' in state_normal.attrib:
            context.error(button,
                          "Button with image '{}' and no title; "
                          "should either have an accessibility label or 'Accessibility Enabled' unchecked",
                          state_normal.attrib['image'])

        if 'backgroundImage' in state_normal.attrib:
            context.error(button,
                          "Button with background image '{}' and no title "
                          "should either have an accessibility label or 'Accessibility Enabled' unchecked",
                          state_normal.attrib['backgroundImage'])

    # For barButtonItem, we expect use of a LyftUI extension
    for bar_button_item in context.tree.findall(".//barButtonItem"):
        if (
            'image' in bar_button_item.attrib and
            'title' not in bar_button_item.attrib and
            get_view_user_defined_attr(bar_button_item, 'voiceoverLabel') is None
        ):
            context.error(bar_button_item,
                          "Bar button item with image {} and no title "
                          "should have a user-defined `voiceoverLabel` attribute", bar_button_item.get('image'))



from setuptools import setup, find_packages
import xiblint

setup(
    name='xiblint',
    version=xiblint.__version__,
    description='Checks .xib and .storyboard files for compliance with best practices',
    license='apache2',
    url='https://github.com/lyft/xiblint',
    author='Ilya Konstantinov',
    author_email='ikonstantinov@lyft.com',
    install_requires=[
        'defusedxml>=0.5.0',
    ],
    packages=find_packages(),
    entry_points={'console_scripts': ['xiblint=xiblint.__main__:main']},
)
